/*
Bullet Continuous Collision Detection and Physics Library
Copyright (c) 2003-2006 Erwin Coumans  http://continuousphysics.com/Bullet/

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the use of this software.
Permission is granted to anyone to use this software for any purpose, 
including commercial applications, and to alter it and redistribute it freely, 
subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/




#include "btBulletDynamicsCommon.h"
#include "LinearMath/btIDebugDraw.h"

#include "GLDebugDrawer.h"

#include "GLDebugFont.h"
#include <stdio.h> //printf debugging

#include "DoorDemo.h"
#include "GL_ShapeDrawer.h"
#include "GlutStuff.h"

#include "GLDebugDrawer.h"
static GLDebugDrawer	gDebugDrawer;



const int numObjects = 3;

#define ENABLE_ALL_DEMOS 1

#define CUBE_HALF_EXTENTS 1.f

#define SIMD_PI_2 ((SIMD_PI)*0.5f)
#define SIMD_PI_4 ((SIMD_PI)*0.25f)




btTransform sliderTransform;
btVector3 lowerSliderLimit = btVector3(-10,0,0);
btVector3 hiSliderLimit = btVector3(10,0,0);

btRigidBody* d6body0 =0;

btHingeConstraint* spDoorHinge = NULL;
btHingeConstraint* spHingeDynAB = NULL;
btGeneric6DofConstraint* spSlider6Dof = NULL;

static bool s_bTestConeTwistMotor = false;



void	drawLimit()
{
		btVector3 from = sliderTransform*lowerSliderLimit;
		btVector3 to = sliderTransform*hiSliderLimit;
		btVector3 color(255,0,0);
		glBegin(GL_LINES);
		glColor3f(color.getX(), color.getY(), color.getZ());
		glVertex3d(from.getX(), from.getY(), from.getZ());
		glVertex3d(to.getX(), to.getY(), to.getZ());
		if (d6body0)
		{
			from = d6body0->getWorldTransform().getOrigin();
			to = from + d6body0->getWorldTransform().getBasis() * btVector3(0,0,10);
			glVertex3d(from.getX(), from.getY(), from.getZ());
			glVertex3d(to.getX(), to.getY(), to.getZ());
		}
		glEnd();
}


void	DoorDemo::setupEmptyDynamicsWorld()
{
	m_collisionConfiguration = new btDefaultCollisionConfiguration();
	m_dispatcher = new btCollisionDispatcher(m_collisionConfiguration);
	m_overlappingPairCache = new btDbvtBroadphase();
	m_constraintSolver = new btSequentialImpulseConstraintSolver();
	m_dynamicsWorld = new btDiscreteDynamicsWorld(m_dispatcher,m_overlappingPairCache,m_constraintSolver,m_collisionConfiguration);

}

void	DoorDemo::clientResetScene()
{
	exitPhysics();
	initPhysics();
}
void	DoorDemo::initPhysics()
{
	setTexturing(true);
	setShadows(true);

	setCameraDistance(26.f);
	m_Time = 0;

	setupEmptyDynamicsWorld();

    m_dynamicsWorld->setDebugDrawer(&gDebugDrawer);
    m_debugMode |= btIDebugDraw::DBG_DrawConstraints|btIDebugDraw::DBG_DrawConstraintLimits|btIDebugDraw::DBG_NoHelpText;


    btCollisionShape* groundShape = new btBoxShape(btVector3(200,2,200));

	m_collisionShapes.push_back(groundShape);
	btTransform groundTransform;
	groundTransform.setIdentity();
    groundTransform.setOrigin(btVector3(0,-1,0));
	btRigidBody* groundBody;
	groundBody= localCreateRigidBody(0, groundTransform, groundShape);
    groundBody->setFriction(0);groundBody->setRestitution(0);


    // Door creation Here:=======================================================================================================

    // TWEAKABLE DEFINITIONS ----------------------------------------------------------------------------------------------------
    #define USE_SLIDER_CONSTRAINTS          // instead of btHingeConstraints
    #define USE_NON_ALIGNED_TRANSFORMS      // This is because in this case I was not able to fix btHingeConstraint angles (they seem to be global axis aligned: maybe we should use the two body ctr with two reference frames to fix this case...)
    // --------------------------------------------------------------------------------------------------------------------------

    #ifndef USE_SLIDER_CONSTRAINTS
    btAlignedObjectArray<btHingeConstraint*> doorConstraints;
    #else //USE_SLIDER_CONSTRAINTS
    btAlignedObjectArray<btSliderConstraint*> doorConstraints;
    #endif //USE_SLIDER_CONSTRAINTS

    // This offsetT offsets ALL doors!
    const btTransform offsetT(
                #ifndef USE_NON_ALIGNED_TRANSFORMS
                btTransform::getIdentity()
                #else //USE_NON_ALIGNED_TRANSFORMS
                btQuaternion(btVector3(0,1,0),btRadians(32)),btVector3(0,0,0)   // Broken with USE_SLIDER_CONSTRAINTS commented out (Why? TO FIX!)
                #endif //USE_NON_ALIGNED_TRANSFORMS
                );   // We can use this to offset all the doors
    const btScalar doorHeight = 10;                  // Meters
    const btVector3 doorAABB(8,doorHeight,0.25);      // both door parts are included

    //-- Increase to fix stability at hinge limits ------------------------
    const btScalar singleDoorMass = 2;              // Kilos------------------->  Tweaking this value from the "default" can improve stiffness considerably (valid for ALL constraints), but as a drawback it alters the dynamic properties of the rigid body.
    const btScalar localInertiaMultiplier = 100;    // ------------------------>  Same as above, but I prefer using this because it does not affect gravity directly (as a drawback it alters the dynamic property of the body too: i.e. it's not "for free").
    //---------------------------------------------------------------------

    localCreateDoor(doorConstraints,offsetT*btTransform(btMatrix3x3::getIdentity(),btVector3(-10,doorHeight*0.5,-15)),doorAABB,singleDoorMass*2,true,localInertiaMultiplier);
    localCreateDoor(doorConstraints,offsetT*btTransform(btMatrix3x3::getIdentity(),btVector3(-0,doorHeight*0.5,-15)),btVector3(doorAABB.x()*0.5,doorAABB.y(),doorAABB.z()),singleDoorMass,false,localInertiaMultiplier);
    localCreateDoor(doorConstraints,offsetT*btTransform(btMatrix3x3::getIdentity(),btVector3(10,doorHeight*0.5,-15)),doorAABB,singleDoorMass*2,true,localInertiaMultiplier);

    //============================================================================================================================



}

// Uses btHingeConstraints: ATM can't handle non-axis-aligned transforms (but AFAICR I've got a project somewhere in which I was able to overcome this issue...)
void DoorDemo::localCreateDoor(btAlignedObjectArray<btHingeConstraint*>& resultOut,const btTransform& T,const btVector3& fullAABBExtents,const btScalar totalMass,bool isDoubleDoors,btScalar inertiaMultFactor,btScalar loLimit,btScalar hiLimit,bool clearResultOutBeforeUsage)
{
    btCollisionShape *cs(NULL);
    btRigidBody *rbA(NULL),*rbB(NULL);

    if (clearResultOutBeforeUsage) resultOut.clear();
    if (isDoubleDoors) {
        const btVector3 hAABB(fullAABBExtents.x()*0.25,fullAABBExtents.y()*0.5,fullAABBExtents.z()*0.5);    // half AABB extents of EACH of the two door parts
        const btScalar mass = totalMass*0.5;                                                                // mass of EACH of the two door parts
        cs = new btBoxShape(hAABB);
        btVector3 localInertia;
        cs->calculateLocalInertia(mass,localInertia);
        localInertia*=inertiaMultFactor;
        localInertia[0]=localInertia[2]=0;

        rbA = localCreateRigidBody(mass,btTransform(T.getBasis(),T.getOrigin()+T.getBasis().getColumn(0)*(hAABB.x())),cs,1,localInertia);
        //rbA->setLinearFactor(btVector3(0,0,0));
        rbA->setFriction(0);
        rbA->setRestitution(0);
        rbB = localCreateRigidBody(mass,btTransform(T.getBasis(),T.getOrigin()-T.getBasis().getColumn(0)*(hAABB.x())),cs,1,localInertia);
        //rbB->setLinearFactor(btVector3(0,0,0));
        rbB->setFriction(0);
        rbB->setRestitution(0);

        btHingeConstraint* cstA = new btHingeConstraint(*rbA,btVector3(hAABB.x(),0,0),btVector3(0,1,0),false);
        cstA->setLimit(loLimit,hiLimit);
        m_dynamicsWorld->addConstraint(cstA,true);
        resultOut.push_back(cstA);

        btHingeConstraint* cstB = new btHingeConstraint(*rbB,btVector3(-hAABB.x(),0,0),btVector3(0,-1,0),false);    // reversing the axis ensures we can use the same loLimit and hiLimit for both cases (good for motor usage)
        cstB->setLimit(loLimit,hiLimit);
        m_dynamicsWorld->addConstraint(cstB,true);
        resultOut.push_back(cstB);
    }
    else {
        const btVector3 hAABB(fullAABBExtents*0.5);    // half AABB extents of the single door part
        const btScalar mass = totalMass;                // mass of EACH of the single door part
        cs = new btBoxShape(hAABB);
        btVector3 localInertia;
        cs->calculateLocalInertia(mass,localInertia);
        localInertia*=inertiaMultFactor;
        localInertia[0]=localInertia[2]=0;

        rbA = localCreateRigidBody(mass,T,cs,1,localInertia);
        //rbA->setLinearFactor(btVector3(0,0,0));

        btHingeConstraint* cstA = new btHingeConstraint(*rbA,btVector3(-hAABB.x(),0,0),btVector3(0,-1,0),false);
        cstA->setLimit(loLimit,hiLimit);
        m_dynamicsWorld->addConstraint(cstA,true);
        resultOut.push_back(cstA);
    }

}

// Uses btSliderConstraints: works in all cases.
void DoorDemo::localCreateDoor(btAlignedObjectArray<btSliderConstraint*>& resultOut,const btTransform& T,const btVector3& fullAABBExtents,const btScalar totalMass,bool isDoubleDoors,btScalar inertiaMultFactor,btScalar loLimit,btScalar hiLimit,bool clearResultOutBeforeUsage)
{
    btCollisionShape *cs(NULL);
    btRigidBody *rbA(NULL),*rbB(NULL);

    if (clearResultOutBeforeUsage) resultOut.clear();
    if (isDoubleDoors) {
        const btVector3 hAABB(fullAABBExtents.x()*0.25,fullAABBExtents.y()*0.5,fullAABBExtents.z()*0.5);    // half AABB extents of EACH of the two door parts
        const btScalar mass = totalMass*0.5;                                                                // mass of EACH of the two door parts
        cs = new btBoxShape(hAABB);
        btVector3 localInertia;
        cs->calculateLocalInertia(mass,localInertia);
        localInertia*=inertiaMultFactor;
        localInertia[0]=localInertia[2]=0;

        rbA = localCreateRigidBody(mass,btTransform(T.getBasis(),T.getOrigin()+T.getBasis().getColumn(0)*(hAABB.x())),cs,1,localInertia);
        //rbA->setLinearFactor(btVector3(0,0,0));
        rbA->setFriction(0);
        rbA->setRestitution(0);
        rbB = localCreateRigidBody(mass,btTransform(T.getBasis(),T.getOrigin()-T.getBasis().getColumn(0)*(hAABB.x())),cs,1,localInertia);
        //rbB->setLinearFactor(btVector3(0,0,0));
        rbB->setFriction(0);
        rbB->setRestitution(0);


        static btAlignedObjectArray<btRigidBody*> staticRbArray;    //TODO: delete it on exiting
        staticRbArray.push_back(new btRigidBody(0,NULL,NULL));
        btRigidBody* rbS = staticRbArray[staticRbArray.size()-1];


        const btMatrix3x3 mSliderToY(btQuaternion(btVector3(0,0,1),SIMD_HALF_PI));

        const btTransform frameA(mSliderToY,btVector3(hAABB.x(),0,0));
        btSliderConstraint* cstA = new btSliderConstraint(*rbA,frameA,false);
        cstA->setLowerAngLimit(loLimit);
        cstA->setUpperAngLimit(hiLimit);
        cstA->setLowerLinLimit(0);
        cstA->setUpperLinLimit(0);
        m_dynamicsWorld->addConstraint(cstA,true);
        resultOut.push_back(cstA);

        const btMatrix3x3 mSliderToMinusY(btQuaternion(btVector3(0,0,1),-SIMD_HALF_PI));     // reversing the axis ensures we can use the same loLimit and hiLimit for both cases (good for motor usage)
        const btTransform frameB(mSliderToMinusY,btVector3(-hAABB.x(),0,0));
        btSliderConstraint* cstB = new btSliderConstraint(*rbB,frameB,false);
        cstB->setLowerAngLimit(loLimit);
        cstB->setUpperAngLimit(hiLimit);
        cstB->setLowerLinLimit(0);
        cstB->setUpperLinLimit(0);
        m_dynamicsWorld->addConstraint(cstB,true);
        resultOut.push_back(cstB);
    }
    else {
        const btVector3 hAABB(fullAABBExtents*0.5);    // half AABB extents of the single door part
        const btScalar mass = totalMass;                // mass of EACH of the single door part
        cs = new btBoxShape(hAABB);
        btVector3 localInertia;
        cs->calculateLocalInertia(mass,localInertia);
        localInertia*=inertiaMultFactor;
        localInertia[0]=localInertia[2]=0;

        rbA = localCreateRigidBody(mass,T,cs,1,localInertia);
        //rbA->setLinearFactor(btVector3(0,0,0));

        const btMatrix3x3 mSliderToMinusY(btQuaternion(btVector3(0,0,1),-SIMD_HALF_PI));

        const btTransform frameA(mSliderToMinusY,btVector3(-hAABB.x(),0,0));
        btSliderConstraint* cstA = new btSliderConstraint(*rbA,frameA,false);
        cstA->setLowerAngLimit(loLimit);
        cstA->setUpperAngLimit(hiLimit);
        cstA->setLowerLinLimit(0);
        cstA->setUpperLinLimit(0);
        m_dynamicsWorld->addConstraint(cstA,true);
        resultOut.push_back(cstA);
    }

}


btRigidBody* DoorDemo::localCreateRigidBody(btScalar mass, const btTransform& startTransform,btCollisionShape* shape,const btScalar localInertiaMultiplier,const btVector3& _localInertia)
{
    btAssert((!shape || shape->getShapeType() != INVALID_SHAPE_PROXYTYPE));

    //rigidbody is dynamic if and only if mass is non zero, otherwise static
    bool isDynamic = (mass != 0.f);

    btVector3 localInertia(0,0,0);
    if (isDynamic)  {
        if (_localInertia.isZero()) shape->calculateLocalInertia(mass,localInertia);
        else localInertia = _localInertia;
        localInertia*=localInertiaMultiplier;
    }

    //using motionstate is recommended, it provides interpolation capabilities, and only synchronizes 'active' objects

#define USE_MOTIONSTATE 1
#ifdef USE_MOTIONSTATE
    btDefaultMotionState* myMotionState = new btDefaultMotionState(startTransform);

    btRigidBody::btRigidBodyConstructionInfo cInfo(mass,myMotionState,shape,localInertia);

    btRigidBody* body = new btRigidBody(cInfo);
    body->setContactProcessingThreshold(m_defaultContactProcessingThreshold);

#else
    btRigidBody* body = new btRigidBody(mass,0,shape,localInertia);
    body->setWorldTransform(startTransform);
#endif//

    m_dynamicsWorld->addRigidBody(body);

    return body;
}

void	DoorDemo::exitPhysics()
{

		int i;

	//removed/delete constraints
	for (i=m_dynamicsWorld->getNumConstraints()-1; i>=0 ;i--)
	{
		btTypedConstraint* constraint = m_dynamicsWorld->getConstraint(i);
		m_dynamicsWorld->removeConstraint(constraint);
		delete constraint;
	}

	//remove the rigidbodies from the dynamics world and delete them
	for (i=m_dynamicsWorld->getNumCollisionObjects()-1; i>=0 ;i--)
	{
		btCollisionObject* obj = m_dynamicsWorld->getCollisionObjectArray()[i];
		btRigidBody* body = btRigidBody::upcast(obj);
		if (body && body->getMotionState())
		{
			delete body->getMotionState();
		}
		m_dynamicsWorld->removeCollisionObject( obj );
		delete obj;
	}




	//delete collision shapes
	for (int j=0;j<m_collisionShapes.size();j++)
	{
		btCollisionShape* shape = m_collisionShapes[j];
		delete shape;
	}

	m_collisionShapes.clear();

	//delete dynamics world
	delete m_dynamicsWorld;

	//delete solver
	delete m_constraintSolver;

	//delete broadphase
	delete m_overlappingPairCache;

	//delete dispatcher
	delete m_dispatcher;

	delete m_collisionConfiguration;

}

DoorDemo::~DoorDemo()
{
	//cleanup in the reverse order of creation/initialization

	exitPhysics();

}


void DoorDemo::clientMoveAndDisplay()
{
	
 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 

 	float dt = float(getDeltaTimeMicroseconds()) * 0.000001f;
	//printf("dt = %f: ",dt);

	// drive cone-twist motor
	m_Time += 0.03f;



	
	{
	 	//during idle mode, just run 1 simulation step maximum
		int maxSimSubSteps = m_idle ? 1 : 1;
		if (m_idle)
			dt = 1.0f/420.f;

		int numSimSteps = m_dynamicsWorld->stepSimulation(dt,maxSimSubSteps);

		//optional but useful: debug drawing
		m_dynamicsWorld->debugDrawWorld();
	
		bool verbose = false;
		if (verbose)
		{
			if (!numSimSteps)
				printf("Interpolated transforms\n");
			else
			{
				if (numSimSteps > maxSimSubSteps)
				{
					//detect dropping frames
					printf("Dropped (%i) simulation steps out of %i\n",numSimSteps - maxSimSubSteps,numSimSteps);
				} else
				{
					printf("Simulated (%i) steps\n",numSimSteps);
				}
			}
		}
	}
	renderme();

//	drawLimit();

    glFlush();
    swapBuffers();
}




void DoorDemo::displayCallback(void) {

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 

	if (m_dynamicsWorld)
		m_dynamicsWorld->debugDrawWorld();

//	drawLimit();

	renderme();

    glFlush();
    swapBuffers();
}


void DoorDemo::keyboardCallback(unsigned char key, int x, int y)
{
	(void)x;
	(void)y;
	switch (key) 
	{
		case 'O' :
			{
				bool offectOnOff;
				if(spDoorHinge)
				{
					offectOnOff = spDoorHinge->getUseFrameOffset();
					offectOnOff = !offectOnOff;
					spDoorHinge->setUseFrameOffset(offectOnOff);
					printf("DoorHinge %s frame offset\n", offectOnOff ? "uses" : "does not use");
				}
				if(spHingeDynAB)
				{
					offectOnOff = spHingeDynAB->getUseFrameOffset();
					offectOnOff = !offectOnOff;
					spHingeDynAB->setUseFrameOffset(offectOnOff);
					printf("HingeDynAB %s frame offset\n", offectOnOff ? "uses" : "does not use");
				}
				if(spSlider6Dof)
				{
					offectOnOff = spSlider6Dof->getUseFrameOffset();
					offectOnOff = !offectOnOff;
					spSlider6Dof->setUseFrameOffset(offectOnOff);
					printf("Slider6Dof %s frame offset\n", offectOnOff ? "uses" : "does not use");
				}
			}
			break;
		default : 
			{
				DemoApplication::keyboardCallback(key, x, y);
			}
			break;
	}
}
